package test;

import com.google.protobuf.ByteString;
import io.grpc.stub.ClientCallStreamObserver;
import io.grpc.stub.ClientResponseObserver;
import io.grpc.stub.StreamObserver;
import io.grpc.stub.StreamObservers;
import io.grpc.testing.GrpcServerRule;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Iterator;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import org.junit.Rule;
import org.junit.Test;

public class UploadTest {

  @Rule
  public final GrpcServerRule grpc = new GrpcServerRule().directExecutor();

  @Test
  public void upload() throws ExecutionException, InterruptedException {
    var result = new CompletableFuture<Boolean>();
    grpc.getServiceRegistry().addService(new TestService());
    var client = UploadServiceGrpc.newStub(grpc.getChannel());

    var data = new TestInputStream(200_000_000);
    var observer = client.upload(new ClientResponseObserver<UploadData, UploadResult>() {
      @Override
      public void beforeStart(ClientCallStreamObserver<UploadData> clientCallStreamObserver) {
        var chunker = new Iterable<UploadData>() {

          // Split `data` InputStream into chunks of `UploadData`
          @Override
          public Iterator<UploadData> iterator() {
            return new Iterator<>() {
              final ByteBuffer buffer = ByteBuffer.allocate(1_000_000);
              boolean done = false;

              @Override
              public boolean hasNext() {
                return !done;
              }

              @Override
              public UploadData next() {
                buffer.clear();

                // try to completely fill the buffer first before sending
                try {
                  do {
                    var bytesRead = data.read(buffer.array(), buffer.position(),
                        buffer.capacity() - buffer.position());
                    done = bytesRead == -1;
                    if (!done) {
                      buffer.position(buffer.position() + bytesRead);
                    }
                  } while (!done && buffer.hasRemaining());
                } catch (IOException e) {
                  throw new RuntimeException(e);
                }

                ByteString byteString = ByteString.EMPTY;
                if (buffer.position() > 0) {
                  buffer.flip();
                  System.out.println("produce chunk. size=" + buffer.remaining());
                  byteString = ByteString.copyFrom(buffer);
                }

                return UploadData.newBuilder()
                    .setData(byteString)
                    .build();
              }
            };
          }
        };

        StreamObservers.copyWithFlowControl(chunker, clientCallStreamObserver);
      }

      @Override
      public void onNext(UploadResult o) {
      }

      @Override
      public void onError(Throwable throwable) {
        result.completeExceptionally(throwable);
      }

      @Override
      public void onCompleted() {
        result.complete(true);
      }
    });

    result.get();
  }

  private static class TestService extends UploadServiceGrpc.UploadServiceImplBase {

    @Override
    public StreamObserver<UploadData> upload(final StreamObserver<UploadResult> responseObserver) {
      return new StreamObserver<>() {
        long bytes = 0L;

        @Override
        public void onNext(final UploadData chunk) {
          bytes += chunk.getData().size();
          System.out.println("got chunk. current size=" + bytes);
        }

        @Override
        public void onError(final Throwable t) {
          System.out.println("Server error: " + t.getMessage());
        }

        @Override
        public void onCompleted() {
          responseObserver.onNext(UploadResult.getDefaultInstance());
          responseObserver.onCompleted();
        }
      };
    }
  }

  private static class TestInputStream extends InputStream {

    public final long sizeInBytes;
    private long pos = 0;

    TestInputStream(final long size) {
      this.sizeInBytes = size;
    }

    public int read() {
      int next = pos < sizeInBytes ? 0 : -1;
      pos++;
      return next;
    }

    @Override
    public void close() {
    }
  }
}
